class NotFound(Exception):
    pass


class AlreadyExists(Exception):
    pass


class InvalidArgumentType(Exception):
    pass


class InvalidClusterName(Exception):
    pass


class InvalidOperator(Exception):
    pass
